import { CONVERTERS, ELEMENT_CACHE, NO_TAG, NO_HTML, NO_CSS, DEFAULT_SHADOWMODE, ELEMENTS } from './Constants';
import global from 'whenthough';
import CacheItem from './CacheItem';
import Presenter from './Presenter';
class UI extends HTMLElement {
    constructor() {
        super();
        this.global = global;
        this.class = this.constructor;
        this.root = this.attachShadow(this.class['mode']);
        this.root.appendChild(this.class.cache.template.content.cloneNode(true));
        this.root.appendChild(this.class.cache.styles.content.cloneNode(true));
        this.class.mixins.forEach(plugin => plugin.connect(this));
    }
    // required shadow element members
    static get observedAttributes() {
        return this.cache.observed;
    }
    connectedCallback() {
        this.dispatch('connected', this);
    }
    adoptedCallback() {
        this.dispatch('adopted', this);
    }
    disconnectedCallback() {
        this.dispatch('disconnected', this);
        this.destructor();
    }
    attributeChangedCallback(name, oldValue, newValue) {
        if (oldValue === newValue) {
            return;
        }
        const properties = this.class.cache.properties;
        const converters = this.class.cache.converters;
        const converter = (converters.get(properties.get(name) || 'string') || (v => v));
        const _old = this[name];
        const _new = converter(newValue);
        this[name] = _new;
        this.dispatch(`${name}Changed`, { old: _old, new: _new });
    }
    // element instance properties
    get presenter() {
        return this.getAttribute('presenter') || '';
    }
    set presenter(value) {
        value ? this.setAttribute('presenter', value) : this.removeAttribute('presenter');
        Presenter.removeFrom(this);
        Presenter.addTo(this, value);
    }
    dispatch(eventName, data) {
        return this.dispatchEvent(new CustomEvent(eventName, { detail: data, bubbles: true, cancelable: true }));
    }
    listen(eventName, listener) {
        this.addEventListener(eventName, listener);
        return this;
    }
    unlisten(eventName, listener) {
        this.removeEventListener(eventName, listener);
        return this;
    }
    select(selector) {
        return Array.prototype.slice.call(this.root.querySelectorAll(selector));
    }
    part(partName) {
        return this.root.querySelector(`[part='${partName}']`);
    }
    // special members
    static get cache() { return ELEMENT_CACHE.get(this.tag); }
    static register() {
        // basic validation
        if (this.tag == NO_TAG) {
            throw new Error('static tag: string, must be redefined.\nString must be a valid custom element tag.');
        }
        if (this.html == NO_HTML) {
            throw new Error('static html: string, must be redefined.');
        }
        if (this.css == NO_CSS) {
            throw new Error('static css: string, must be redefined.');
        }
        // register
        if (this.cache) {
            return;
        }
        // template
        const template = document.createElement('template');
        template.innerHTML = this.html;
        // styles
        const styles = document.createElement('template');
        let css = this.css;
        if (typeof css === 'string') {
            css = [css];
        }
        if (Array.isArray(css)) {
            styles.innerHTML = css
                .map(s => `<style type="text/css">\n${s}\n</style>`)
                .join('\n');
        }
        // observed props
        const observed = this.observed;
        // built-in props
        observed.set('presenter', 'string');
        const properties = [...observed.keys()];
        const cache = new CacheItem(this.tag, template, styles, this, this.converters, this.observed, properties);
        ELEMENT_CACHE.set(this.tag, cache);
        ELEMENTS.set(this.tag, this);
        this.mixins.forEach(plugin => plugin.attach(this));
        customElements.define(this.tag, this);
        global.set(Symbol.for(`@wayang/element/${this.tag}`), this);
    }
    destructor() {
        Presenter.removeFrom(this);
        this.class.cache.mixins.forEach(plugin => plugin.disconnect(this));
    }
}
// must be re-defined
UI.tag = NO_TAG;
UI.html = NO_HTML;
UI.css = NO_CSS;
UI.observed = new Map();
// optionally re-defined
UI.converters = new Map(CONVERTERS);
UI.mode = DEFAULT_SHADOWMODE;
UI.mixins = new Array();
export default UI;
