import global from 'whenthough';
import { CONVERTERS, COMPONENT_CACHE, NO_TAG, NO_HTML, NO_CSS, DEFAULT_SHADOWMODE, COMPONENTS } from './Constants';
import CacheItem from './CacheItem';
import Presenter from './Presenter';
const debounceKey = Symbol.for('@wayang/web-component/attributeChangedCallback/debounce');
class WebComponent extends HTMLElement {
    constructor() {
        super();
        this.global = global;
        this.debouncers = {};
        this.class = this.constructor;
        this.root = this.attachShadow(this.class.mode);
        this.root.appendChild(this.class.cache.template.content.cloneNode(true));
        this.root.appendChild(this.class.cache.styles.content.cloneNode(true));
        this.class.mixins.forEach(mixin => mixin.connect(this));
    }
    // required shadow element members
    static get observedAttributes() {
        return this.cache.observed;
    }
    connectedCallback() {
        this.dispatch('connected', this);
    }
    adoptedCallback() {
        this.dispatch('adopted', this);
    }
    disconnectedCallback() {
        this.dispatch('disconnected', this);
        this.destructor();
    }
    attributeChangedCallback(name, previous, current) {
        if (previous === current) {
            return;
        }
        clearTimeout(this.debouncers);
        this.debouncers = setTimeout(() => {
            const properties = this.class.cache.properties;
            const converters = this.class.cache.converters;
            const converter = (converters.get(properties.get(name) || 'string') || (v => v));
            const _old = this[name];
            const _new = converter(current);
            this[name] = _new;
            this.dispatch(`${name}Changed`, { old: _old, new: _new });
        }, 0);
    }
    // element instance properties
    get presenter() {
        return this.getAttribute('presenter') || '';
    }
    set presenter(value) {
        value ? this.setAttribute('presenter', value) : this.removeAttribute('presenter');
        Presenter.removeFrom(this);
        Presenter.addTo(this, value);
    }
    // element instance methods
    dispatch(eventName, data) {
        return this.dispatchEvent(new CustomEvent(eventName, { detail: data, bubbles: true, cancelable: true }));
    }
    listen(eventName, listener) {
        this.addEventListener(eventName, listener);
        return this;
    }
    unlisten(eventName, listener) {
        this.removeEventListener(eventName, listener);
        return this;
    }
    select(selector) {
        return Array.prototype.slice.call(this.root.querySelectorAll(selector));
    }
    // special members
    static get cache() { return COMPONENT_CACHE.get(this.tag); }
    static register() {
        // basic validation
        if (this.tag == NO_TAG) {
            throw new Error('static tag: string, must be redefined.\nString must be a valid custom element tag.');
        }
        if (this.html == NO_HTML) {
            throw new Error('static html: string, must be redefined.');
        }
        if (this.css == NO_CSS) {
            throw new Error('static css: string, must be redefined.');
        }
        // register
        if (this.cache) {
            return;
        }
        // template
        const template = document.createElement('template');
        template.innerHTML = this.html;
        // styles
        const styles = document.createElement('template');
        let css = this.css;
        if (typeof css === 'string') {
            css = [css];
        }
        if (Array.isArray(css)) {
            styles.innerHTML = css
                .map(s => `<style type="text/css">\n${s}\n</style>`)
                .join('\n');
        }
        // observed props
        const observed = this.observed;
        // built-in props
        observed.set('presenter', 'string');
        const properties = [...observed.keys()];
        const cache = new CacheItem(this.tag, template, styles, this, this.converters, this.observed, properties);
        COMPONENT_CACHE.set(this.tag, cache);
        COMPONENTS.set(this.tag, this);
        this.mixins.forEach(mixin => mixin.attach(this));
        customElements.define(this.tag, this);
        global.set(Symbol.for(`@wayang/web-components/${this.tag}`), this);
    }
    destructor() {
        Presenter.removeFrom(this);
        this.class.cache.mixins.forEach(mixin => {
            mixin.disconnect(this);
            const mixinKey = Symbol.for(`@wayang/mixin/state/${mixin.id}`);
            this[mixinKey] = null;
            delete this[mixinKey];
        });
    }
}
// must be re-defined
WebComponent.tag = NO_TAG;
WebComponent.html = NO_HTML;
WebComponent.css = NO_CSS;
WebComponent.observed = new Map();
// optionally re-defined
WebComponent.converters = new Map(CONVERTERS);
WebComponent.mode = DEFAULT_SHADOWMODE;
WebComponent.mixins = new Array();
export default WebComponent;
